"""
@author  : MG
@Time    : 2020/11/16 10:20
@File    : template.py
@contact : mmmaaaggg@163.com
@desc    : 用于
"""
import json
import math
import threading
import time
from datetime import datetime
from typing import List, Dict, Optional, Tuple

import numpy as np
import pandas as pd
from ibats_utils.decorators import thread_save
from ibats_utils.mess import datetime_2_str
from vnpy.app.portfolio_strategy import StrategyEngine
from vnpy.app.portfolio_strategy import StrategyTemplate as StrategyTemplateBase
from vnpy.trader.constant import Direction, Offset, Interval
from vnpy.trader.object import BarData, TickData, TradeData

from vnpy_extra.backtest import STOP_OPENING_POS_PARAM, ENABLE_COLLECT_DATA_PARAM, StopOpeningPos, \
    check_datetime_trade_available, generate_mock_load_bar_data
from vnpy_extra.backtest import check_datetime_available
from vnpy_extra.config import logging
from vnpy_extra.constants import GeneralPeriodEnum
from vnpy_extra.db.orm import AccountStrategyStatusEnum
from vnpy_extra.report.collector import trade_data_collector, order_data_collector, latest_price_collector
from vnpy_extra.report.monitor import AccountStrategyStatusMonitor
from vnpy_extra.utils.enhancement import BarGenerator, ArrayManager, get_daily_min_bar_count


class StrategyTemplate(StrategyTemplateBase):
    # 该标识位默认为0（关闭状态）。为1时开启，程序一旦平仓后则停止后续交易。该标识位用于在切换合约时使用
    stop_opening_pos = StopOpeningPos.open_available.value
    # 考虑到六、周日、节假日等非交易日因素，保险起见，建议初始化日期 * 2 + 7
    init_load_days = 30
    # 加载主连连续合约作为合约历史行情数据（默认为False)
    load_main_continuous_md = False
    # 对应周期类型
    period_enum = None
    # 周期级别（如果设置 period_enum，则该字段无效）
    interval = None
    # 周期大小（如果设置 period_enum，则该字段无效）
    window = None
    # ArrayManager 队列长度
    array_size = 100

    def __init__(
            self,
            strategy_engine: StrategyEngine,
            strategy_name: str,
            vt_symbols: List[str],
            setting: dict
    ):
        super().__init__(strategy_engine, strategy_name, vt_symbols, setting)
        # setting 不包含 stop_opening_pos key
        self.setting = {k: v for k, v in setting.items() if k in self.parameters}
        self.parameters.append(STOP_OPENING_POS_PARAM)  # 增加 stop_opening_pos 用于合约切换是关闭当前线程
        self.logger = logging.getLogger(f'strategies.portfolio.{strategy_name}')
        # 写日志
        self.logger.info(f"{strategy_name} on {vt_symbols} setting=\n{json.dumps(setting, indent=4)}")
        # 仅用于 on_order 函数记录上一个 order 使用，解决vnpy框架重复发送order的问题
        self._last_order = {}
        self.received_trade_list = []  # 记录所有成交数据
        # 1分钟 bar 对象字典及bar数量
        self.current_bars: Optional[Dict[str, BarData]] = None
        self.bar_count = 0
        # 生成指定周期 bar 对象字典及 win bar 数量
        self.current_win_bars: Optional[Dict[str, BarData]] = None
        self.win_bar_count = 0
        # 是否实盘环境
        self._is_realtime_mode = self.strategy_name is not None and self.strategy_name != self.__class__.__name__
        self._strategy_status = AccountStrategyStatusEnum.Created
        # 是否收集申请单以及交易单记录
        self.enable_collect_data = ENABLE_COLLECT_DATA_PARAM in setting and setting[ENABLE_COLLECT_DATA_PARAM]
        self._strategy_status_monitor: Optional[AccountStrategyStatusMonitor] = None
        self._lock: Optional[threading.Lock] = None

        # 最近一个tick的时间
        self.last_tick_time: Optional[datetime] = None
        # 最近一次下单的时间
        self.last_order_dt: Dict[str, datetime] = {}
        # 最近一条提示信息
        self._last_msg = ''

        def on_bar(bar: BarData):
            """"""
            pass

        self.bgs: Dict[str, BarGenerator] = {}
        self.ams: Dict[str, ArrayManager] = {}
        if self.period_enum is not None:
            period_enum: GeneralPeriodEnum = GeneralPeriodEnum[self.period_enum]
            self.window, self.interval = period_enum.value

        self._win_bars = {}
        for vt_symbols in self.vt_symbols:
            self.bgs[vt_symbols] = BarGenerator(
                on_bar, window=self.window, interval=self.interval, on_window_bar=self._on_win_bar)
            self.ams[vt_symbols] = ArrayManager(self.array_size)
            self._win_bars[vt_symbols] = None

    def set_is_realtime_mode(self, is_realtime_mode):
        self._is_realtime_mode = is_realtime_mode

    def get_id_name(self):
        """获取 id_name"""
        param_str = '_'.join([
            str(self.setting[key]) for key in self.parameters
            if key in self.setting and key != STOP_OPENING_POS_PARAM])
        id_name = f"{self.__class__.__name__}_{param_str}" if param_str != "" else f"{self.__class__.__name__}"
        return id_name

    def _set_strategy_status(self, status: AccountStrategyStatusEnum):
        if not self._is_realtime_mode:
            # 仅针对实时交易是使用
            return
        if self._strategy_status == status:
            return

        if status == AccountStrategyStatusEnum.RunPending and self._strategy_status not in (
                AccountStrategyStatusEnum.Created, AccountStrategyStatusEnum.Running
        ):
            # AccountStrategyStatusEnum.RunPending 状态只从数据库端发起
            if self._lock is not None:
                self._lock.acquire()

            try:
                # 保险起见，为防止出现死循环调用，在 on_start 先把状态调整过来
                self._strategy_status = status
                self.write_log(f"策略 {self.strategy_name} {self.vt_symbols} 状态 "
                               f"{self._strategy_status.name} -> {status.name} 被远程启动")
            finally:
                if self._lock is not None:
                    self._lock.release()

            self.on_start()

        elif status == AccountStrategyStatusEnum.StopPending and self._strategy_status == AccountStrategyStatusEnum.Running:
            # AccountStrategyStatusEnum.StopPending 状态只从数据库端发起
            if self._lock is not None:
                self._lock.acquire()

            try:
                # 保险起见，为防止出现死循环调用，在 on_stop 先把状态调整过来
                self._strategy_status = status
                self.write_log(f"策略 {self.strategy_name} {self.vt_symbols} 状态 "
                               f"{self._strategy_status.name} -> {status.name} 被远程停止")
            finally:
                if self._lock is not None:
                    self._lock.release()

            self.on_stop()
        else:
            self.write_log(f"策略 {self.strategy_name} {self.vt_symbols} 状态 "
                           f"{self._strategy_status.name} -> {status.name}")
            self._strategy_status = status

    def _get_strategy_status(self) -> AccountStrategyStatusEnum:
        return self._strategy_status

    def on_init(self) -> None:
        super().on_init()
        self.bar_count = 0
        if self._is_realtime_mode:
            self._set_strategy_status(AccountStrategyStatusEnum.Initialized)

        if self._is_realtime_mode and self._strategy_status_monitor is None:
            # 该语句一定不能放在 __init__ 中
            # 因为 strategy_name 在回测阶段模块中，在 __init__ 后可能会被重写赋值
            self._strategy_status_monitor = AccountStrategyStatusMonitor(
                self.strategy_name,
                self._get_strategy_status,
                self._set_strategy_status,
                self.vt_symbols,
                self.setting
            )
            self._lock = self._strategy_status_monitor.lock

        if self._strategy_status_monitor is not None and not self._strategy_status_monitor.is_alive():
            self._strategy_status_monitor.start()

        self.enable_collect_data |= self._is_realtime_mode
        if self.enable_collect_data:
            trade_data_collector.queue_timeout = 90 if self._is_realtime_mode else 1
            order_data_collector.queue_timeout = 90 if self._is_realtime_mode else 1

        if self._is_realtime_mode and self.load_main_continuous_md:
            with generate_mock_load_bar_data(self.write_log):
                self.load_bars(self.init_load_days)
        else:
            self.load_bars(self.init_load_days)

        self.write_log(f"策略初始化完成. 加载{self.init_load_days}天数据")

    def on_start(self) -> None:
        super().on_start()
        if self._is_realtime_mode:
            self._set_strategy_status(AccountStrategyStatusEnum.Running)
        # 整理持仓信息
        if len(self.pos) == 0:
            self.write_log(f"策略启动，当前初始持仓：(无)")
        else:
            pos_str = '\t'.join([f"{k}: {v}" for k, v in self.pos.items()])
            self.write_log(f"策略启动，当前初始持仓：\n{pos_str}")

        self.put_event()

        # 初始化相关数据,可以在重启策略时清空历史订单对当前策略的影响.
        # 仅用于 on_order 函数记录上一个 order 使用，解决vnpy框架重复发送order的问题
        self._last_order = {}
        self.received_trade_list = []  # 记录所有成交数据
        # 最近一次下单的时间
        self.last_order_dt: Dict[str, datetime] = {}
        # if self._is_realtime_mode:
        #     start_strategy_position_monitor()

    def on_tick(self, tick: TickData) -> bool:
        """判断当前tick数据是否有效,如果无效数据直接返回 False,否则更新相应bar数据"""
        super().on_tick(tick)
        is_available = check_datetime_available(tick.datetime)
        if not is_available:
            return is_available

        if (
                self.last_tick_time
                and self.last_tick_time.minute != tick.datetime.minute
        ):
            bars, all_none = {}, False
            for vt_symbol, bg in self.bgs.items():
                bars[vt_symbol] = bar = bg.generate()
                if bar is not None:
                    all_none = False

            if not all_none:
                self.on_bars(bars)

        bg: BarGenerator = self.bgs[tick.vt_symbol]
        bg.update_tick(tick)
        latest_price_collector.put_nowait(tick)
        self.last_tick_time = tick.datetime.replace(tzinfo=None)
        return is_available

    def on_bars(self, bars: Dict[str, BarData]) -> None:
        super().on_bars(bars)
        self.current_bars: Dict[str, BarData] = bars
        self.bar_count += 1
        for vt_symbol, bar in bars.items():
            if bar is not None:
                self.bgs[vt_symbol].update_bar(bar)

    def _on_win_bar(self, bar: BarData):
        """
        激活 on_win_bars 函数
        逻辑:
        通过 bg 生成对应 vt_symbol 的 win bar,当全部 vt_symbol 的 win bar 均生成完毕后,触发 on_win_bars 函数
        由于 on_tick 阶段 通过 gb.generate() 强行在同一时刻生成了所有 bar,因此可以保证,当生成对应 win bar 时,也会是同时生成,
        因此,此处不考虑部分生成的情况
        """
        self._win_bars[bar.vt_symbol] = bar
        bar_dt = None
        for bar in list(self._win_bars.values()):
            if bar is None:
                break
            if bar_dt is None:
                bar_dt = bar.datetime
            elif bar_dt != bar.datetime:
                # 判断bar时间是否同步
                break
        else:
            ret = self.on_win_bars(self._win_bars)
            if ret is None or ret:
                self.on_win_bars_finished(self._win_bars)
            else:
                self.on_win_bars_failed(self._win_bars)

            self._win_bars = {_: None for _ in self._win_bars.keys()}

    def on_win_bars(self, bars: Dict[str, BarData]) -> Optional[bool]:
        self.current_win_bars: Dict[str, BarData] = bars
        self.win_bar_count += 1
        for vt_symbols, bar in bars.items():
            self.ams[vt_symbols].update_bar(bar)

        return True

    def on_win_bars_finished(self, bars: Dict[str, BarData]) -> None:
        """对 on_win_bars 逻辑进行收尾处理"""
        pass

    def on_win_bars_failed(self, bars: Dict[str, BarData]) -> None:
        """对 on_win_bars 逻辑进行异常处理"""
        pass

    def send_order(self,
                   vt_symbol: str,
                   direction: Direction,
                   offset: Offset,
                   price: float,
                   volume: float,
                   lock: bool = False
                   ) -> List[str]:
        current_pos = int(self.get_pos(vt_symbol))
        order_datetime = self.current_bars[vt_symbol].datetime if (
                self.current_bars is not None
                and vt_symbol in self.current_bars
                and self.current_bars[vt_symbol] is not None) else None
        ignore_order = False
        if price <= 0.0:
            log_level = 'error'
            msg = '【价格无效】'
            ignore_order = True
        elif order_datetime is not None and not check_datetime_trade_available(order_datetime):
            # 回测模式下提示此类信息
            log_level = 'warning' if self._is_realtime_mode else 'debug'
            msg = '【非交易时段】'
            ignore_order = True
        elif self.stop_opening_pos != StopOpeningPos.open_available.value and offset == Offset.OPEN:
            log_level = 'warning'
            msg = '【禁止开仓】'
            ignore_order = True
        else:
            log_level = 'debug'
            msg = ''

        if offset == Offset.OPEN:
            if self.stop_opening_pos != StopOpeningPos.stop_opening_and_nolog.value:
                self.write_log(
                    f"{vt_symbol:>11s} {direction.value} {offset.value:4s} {price:.1f} "
                    f"{current_pos:+d} {volume:+.0f} "
                    if order_datetime is None else
                    f"{datetime_2_str(order_datetime)} {vt_symbol} {direction.value} {offset.value:4s} {price:.1f} "
                    f"{current_pos:+d} {volume:+.0f}{msg}",
                    log_level
                )

            if self.stop_opening_pos != StopOpeningPos.open_available.value:
                self.write_log(f"当前策略 stop_opening_pos="
                               f"{self.stop_opening_pos}<{StopOpeningPos(self.stop_opening_pos).name}>，"
                               f"所有开仓操作将被屏蔽（用于主力合约切换,或关闭失效策略使用）", log_level)
                self.stop_opening_pos = StopOpeningPos.stop_opening_and_nolog.value

        else:
            self.write_log(
                f"{vt_symbol:>11s} {direction.value} {offset.value:4s} {price:.1f} "
                f"      {current_pos:+d} {-volume:+.0f} "
                if order_datetime is None else
                f"{datetime_2_str(order_datetime)} {vt_symbol} {direction.value} {offset.value:4s} {price:.1f} "
                f"      {current_pos:+d} {-volume:+.0f}{msg}",
                log_level
            )

        if order_datetime is None:
            order_datetime = datetime.now()

        if ignore_order:
            return []

        if vt_symbol in self._last_order:
            # 记录订单情况，跳过重复订单
            last_order = self._last_order[vt_symbol]
            if last_order['direction'] != direction.value \
                    or last_order['offset'] != offset.value \
                    or last_order['price'] != price \
                    or last_order['volume'] != volume:
                is_not_duplicate_order = True
            else:
                is_not_duplicate_order = False
        else:
            is_not_duplicate_order = True

        if is_not_duplicate_order:
            symbol, exchange = vt_symbol.split('.')
            order = {
                "datetime": order_datetime,
                "symbol": symbol,
                "exchange": exchange,
                "direction": direction.value,
                "offset": offset.value,
                "price": price,
                "volume": volume,
            }
            if self.enable_collect_data:
                order_data_collector.put_nowait(self.strategy_name, order)

            self._last_order[vt_symbol] = order

        self.last_order_dt[vt_symbol] = datetime.now()
        return super().send_order(vt_symbol, direction, offset, price, volume, lock)

    def on_stop(self):
        super().on_stop()
        if self._is_realtime_mode:
            self._set_strategy_status(AccountStrategyStatusEnum.Stopped)
        self.put_event()

    def update_trade(self, trade: TradeData):
        super().update_trade(trade)
        self.received_trade_list.append(trade)
        if self.enable_collect_data:
            trade_data_collector.put_nowait(self.strategy_name, trade)

    def write_log(self, msg: str, logger_method='info'):
        if self._last_msg == msg:
            return
        msg = f"{self.strategy_name} {msg}"
        super().write_log(msg)
        getattr(self.logger, logger_method)(msg)
        self._last_msg = msg

    def is_available(self, stats: dict, df: pd.DataFrame) -> bool:
        balance_s = df["balance"]
        is_available = not (
                stats['total_net_pnl'] <= 0
                or stats['daily_trade_count'] < 0.2  # 1/0.2 每一次交易平仓再开仓需要2次交易，因此相当于10天交易一次
                or stats['return_drawdown_ratio'] < 1.5
                or stats['max_new_higher_duration'] >= 180  # 最长不创新高周期<180
                or stats['max_new_higher_duration'] / stats['total_days'] > 0.5  # 最长不创新高周期超过一半的总回测天数
                or balance_s is None  # 没有交易
                or np.sum((balance_s - stats['capital']) <= 0) / stats['total_days'] > 0.5  # 50%以上交易日处于亏损状态
                or np.all(df["drawdown"][df["drawdown"].idxmin():] < 0)  # 最大回撤到最后一个交易日需要出现新高
        )
        return is_available

    def calc_init_load_days(self):
        """返回至少需要多少日线数据"""
        if self.interval == Interval.MINUTE:
            min_bar_count = max([get_daily_min_bar_count(_) for _ in self.vt_symbols])
            step = 1 / min_bar_count
        elif self.interval == Interval.HOUR:
            min_bar_count = max([get_daily_min_bar_count(_) for _ in self.vt_symbols])
            step = 60 / min_bar_count
        elif self.interval == Interval.DAILY:
            step = 1
        elif self.interval == Interval.WEEKLY:
            step = 7
        else:
            raise ValueError(f"interval={self.interval} 无效")

        daily_bar_count = math.ceil(self.window * self.array_size * step)
        if self.interval != Interval.WEEKLY:
            daily_bar_count *= 1.5

        return daily_bar_count


class TargetPosAndPrice2SymbolTemplate(StrategyTemplate):
    # 该参数仅用于回测时使用（0，两腿同时下单，1：两腿分不同的bar下单）
    handle_order_one_by_one = 0
    # 保持最近的 n 个tick 用于算法交易中进行成交难度评估使用, 0.5秒一个tick
    tick_arr_len = 60

    # 算法交易部分配置信息
    # 最大可容忍配对交易不平衡时间（秒）
    mismatch_endure_seconds = 20
    # 主动成交订单等待时长（秒）
    active_order_wait_seconds = 5

    def __init__(
            self,
            strategy_engine: StrategyEngine,
            strategy_name: str,
            vt_symbols: List[str],
            setting: dict
    ):
        """"""
        super().__init__(strategy_engine, strategy_name, vt_symbols, setting)
        self.vt_symbol1 = vt_symbols[0]
        self.vt_symbol2 = vt_symbols[1]
        # 目标仓位
        self.targets: Dict[str, Tuple[int, Optional[float]]] = {}
        # 目标仓位刷新标识
        self.target_refreshed = False
        # 合约对不匹配的起始时间
        self.active_order_waiting_since_dt = None
        # 单腿情况下，首次尝试盘口成交时，使用盘口挂单方式成交（尝试一次不行后，该为盘口对价成交）
        self.use_market_active_order = False
        # 保存最近n个tick 价格及交易量数据用于计算开仓优先级
        # 其中 np.ndarray 是一个 np.zeros((tick_arr_len, 2)) 第一列是 价格，第二列是成交量
        self.tick_arr: Dict[str, np.ndarray] = {}
        # 成交难度系数。下单时，选择难度系数最高的先下单
        self.deal_difficult_ratio: Dict[str, Tuple[float, datetime]] = {}
        for vt_symbol in self.vt_symbols:
            self.targets[vt_symbol] = (0, 0.0)
            # last_price, last_volume
            self.tick_arr[vt_symbol] = np.zeros((self.tick_arr_len, 2))
            self.deal_difficult_ratio[vt_symbol] = (-1, datetime.now())

        # 算法交易线程
        self.algo_trading_thread = None
        self.algo_trading_running = True
        self.seconds_since_last_tick = 0

    @property
    def order_waiting_since(self):
        return None if self.active_order_waiting_since_dt is None else datetime_2_str(
            self.active_order_waiting_since_dt)

    @order_waiting_since.setter
    def order_waiting_since(self, val):
        pass

    @property
    def difficult_ratio(self):
        return {key: _[0] for key, _ in self.deal_difficult_ratio.items()}

    @difficult_ratio.setter
    def difficult_ratio(self, val):
        pass

    @thread_save
    def _handle_active_order(self, vt_symbol):
        """执行主动单"""
        if self.active_order_waiting_since_dt is None:
            self.active_order_waiting_since_dt = datetime.now()
            mismatched_seconds = 0
        else:
            mismatched_seconds = (datetime.now() - self.active_order_waiting_since_dt).seconds

        # 是否超过被动交易可忍受时限
        intolerable = mismatched_seconds > self.mismatch_endure_seconds
        no_relative_order = True
        if len(self.active_orderids) > 0:
            for orderid in self.get_all_active_orderids():
                order = self.get_order(orderid)
                if order is None:
                    continue
                if order.vt_symbol == vt_symbol:
                    no_relative_order = False

        if no_relative_order:
            if intolerable:
                # 超过时限，对价主动成交
                self.write_log(f"超过被动交易时限，对价主动成交")
                self._handle_pos_2_target(vt_symbol, market_price=True, activate=self.use_market_active_order)
                # 单腿情况下，首次尝试盘口成交时，使用盘口挂单方式成交（尝试一次不行后，该为盘口对价成交）
                self.use_market_active_order = True
            else:
                # 限价被动成交
                self.write_log(f"在时限内，限价被动交易")
                self._handle_pos_2_target(vt_symbol)
        else:
            # 单腿情况，下单后等待成交
            # 距离上一次下单时间
            if vt_symbol in self.last_order_dt:
                last_order_seconds = (datetime.now() - self.last_order_dt[vt_symbol]).seconds
                order_wait_timeout = self.active_order_wait_seconds < last_order_seconds
            else:
                order_wait_timeout = True

            if intolerable and order_wait_timeout:
                self.write_log("单腿等待超时，撤单重下")
                # self.cancel_all()
                self.cancel_order_by_vt_symbol(vt_symbol)

    @thread_save
    def _handle_passive_order(self, vt_symbol):
        """执行被动单（同时用于回测及实盘）"""
        if len(self.active_orderids) == 0:
            # 首次开仓
            self.write_log(f"建立配对交易 {vt_symbol} 优先配对。", 'debug')
            self._handle_pos_2_target(vt_symbol)
            self.target_refreshed = False

        else:
            # 有开仓，等待成交，同时又有新的目标仓位更新
            if self.target_refreshed:
                # 如果超时，则重新下单
                # 检查新下单的价格是否与就单子相同，如果相同则不用 撤销重下单，继续等待即可
                # 注意：这里 list(self.active_orderids) 其中 list() 不能去掉，
                #       因为内部会 cancel_order(order.orderid) 造成 active_orderids 变化会抛异常
                for active_orderid in list(self.active_orderids):
                    order = self.get_order(active_orderid)
                    if order is None:
                        continue
                    target_pos, target_price = self.targets[order.vt_symbol]
                    if order.price != target_price:
                        # 不要在订单取消后面直接下单，
                        # 因为 cancel_order 可能会失败，引发导致重复下单的情况。
                        # 订单取消成功后,通过 if len(self.active_orderids) == 0 状态进行重新下单
                        # 重新下单后 self.target_refreshed = False
                        # 发现 cancel_order(order.orderid) 总是找不到订单导致撤单失败
                        # self.cancel_order(order.orderid)
                        self.write_log(f"订单已经被刷星，取消旧订单", 'debug')
                        self.cancel_all()

    def _algo_trading_2_target_pos_thread(self):
        self.write_log("开启算法交易")
        try:
            while self.algo_trading_running:
                time.sleep(0.5)
                if not self.trading:
                    continue

                if self.last_tick_time is None:
                    # 未开盘期间不进行算法交易
                    continue
                now = datetime.now().replace(tzinfo=self.last_tick_time.tzinfo)
                # 多线程情况下可能存在 last_tick_time > now 的情况，因此用绝对值
                self.seconds_since_last_tick = abs(now - self.last_tick_time).seconds
                if self.seconds_since_last_tick > 5:
                    continue
                self._algo_trading_2_target_pos()
        except:
            self.write_log(f"{self.strategy_name} 算法交易异常", 'error')
            self.logger.exception("%s 算法交易异常", self.strategy_name)
        finally:
            self.write_log("关闭算法交易", 'error')

    def _algo_trading_2_target_pos(self):
        # 检查持仓
        # 计算是否存在仓位不匹配情况
        # 首先分别计算 symbol1 symbol2 的仓位差
        vt_symbol = self.vt_symbol1
        target_pos1, target_price1 = self.targets[vt_symbol]
        current_pos1 = self.get_pos(vt_symbol)
        pos_diff1 = target_pos1 - current_pos1
        pos_diff1_abs = abs(pos_diff1)
        vt_symbol = self.vt_symbol2
        target_pos2, target_price2 = self.targets[vt_symbol]
        current_pos2 = self.get_pos(vt_symbol)
        pos_diff2 = target_pos2 - current_pos2
        pos_diff2_abs = abs(pos_diff2)
        if pos_diff1 == pos_diff2 == 0:
            # 没有需要操作的合约或订单
            self.cancel_all()
            self.active_order_waiting_since_dt = None
            self.target_refreshed = False
            # 单腿情况下，首次尝试盘口成交时，使用盘口挂单方式成交（尝试一次不行后，该为盘口对价成交）
            self.use_market_active_order = False
            return

        # 确定哪一条是主动腿，哪一条是被动腿。判断逻辑如下：
        deal_passive_order = True
        # 是否为初始开仓交易阶段
        is_open_pos = current_pos1 == 0 and current_pos2 == 0
        # 是否为初始关仓交易阶段
        is_close_pos = current_pos1 != 0 and current_pos2 != 0
        # 首先做是否单腿判断
        if pos_diff1_abs > pos_diff2_abs and not (is_open_pos or is_close_pos):
            # 如果 leg1 leg2 数量不匹配，即，存在单腿，以瘸腿的一个为主动腿，另一条为被动腿
            # leg_active 主动腿, leg_passive 被动腿
            leg_passive = self.vt_symbol2
            leg_active = self.vt_symbol1
            # 发主动成交申请
            deal_passive_order = False
            # 重置主动单等待时间
            # self.active_order_waiting_since_dt = None
        elif pos_diff1_abs < pos_diff2_abs and not (is_open_pos or is_close_pos):
            leg_passive = self.vt_symbol1
            leg_active = self.vt_symbol2
            # 发主动成交申请
            deal_passive_order = False
            # 重置主动单等待时间
            # self.active_order_waiting_since_dt = None
        else:
            # 初始建仓，或关仓阶段
            # 不存在单腿情况，进行成交评估。
            leg_passive, leg_active = None, None
            if len(self.active_orderids) > 0:
                for orderid in self.get_all_active_orderids():
                    order = self.get_order(orderid)
                    if order is None:
                        continue

                    leg_passive = order.vt_symbol
                    leg_active = self.vt_symbol1 if self.vt_symbol1 != order.vt_symbol else self.vt_symbol2
                    break

            if leg_passive is None:
                # 选择评估时间长的作为被动腿（优先下单），主动腿等被动腿成交后在发交易申请
                leg1_ratio, leg2_ratio = self.estimate_deal_seconds_all()
                if leg1_ratio == leg2_ratio:
                    # 如果难度相同，leg2 先下单
                    leg_passive = self.vt_symbol2
                    leg_active = self.vt_symbol1
                elif leg1_ratio == -1:
                    # leg1 难成交，leg1 先下被动单
                    leg_passive = self.vt_symbol1
                    leg_active = self.vt_symbol2
                elif leg2_ratio == -1:
                    # leg2 难成交，leg2 先下被动单
                    leg_passive = self.vt_symbol2
                    leg_active = self.vt_symbol1
                elif leg1_ratio < leg2_ratio:
                    # leg1 成交耗时比 leg2 短，leg2 先下被动单
                    leg_passive = self.vt_symbol2
                    leg_active = self.vt_symbol1
                else:
                    # leg2 成交耗时比 leg1 短，leg1 先下被动单
                    leg_passive = self.vt_symbol1
                    leg_active = self.vt_symbol2

        if deal_passive_order:
            # 执行被动单
            self._handle_passive_order(leg_passive)
        else:
            # 执行主动单
            self._handle_active_order(leg_active)

    def on_start(self) -> None:
        super().on_start()
        if self._is_realtime_mode:
            self.algo_trading_running = True
            self.algo_trading_thread = threading.Thread(target=self._algo_trading_2_target_pos_thread, daemon=True)
            self.algo_trading_thread.start()

    def on_stop(self):
        super().on_stop()
        if self._is_realtime_mode:
            self.algo_trading_running = False
            self.algo_trading_thread.join()

    def cancel_order_by_vt_symbol(self, vt_symbol) -> bool:
        """
        根据 vt_symbol 取消订单
        return: True 存在订单被取消的动作；False，没有订单被取消的动作发生（说明已经没有相关的订单）
        """
        canceled = False
        for vt_orderid in self.active_orderids.copy():
            if vt_orderid not in self.orders:
                continue
            order = self.orders[vt_orderid]
            # self.write_log(
            #     f"vt_orderid={vt_orderid} "
            #     f"mismatched={mismatched_vt_symbol} order.vt_symbol {order.vt_symbol} "
            #     f"match={order.vt_symbol == mismatched_vt_symbol}")
            if order.vt_symbol == vt_symbol:
                self.write_log(f'Unbalanced，撤单{vt_orderid}[{vt_symbol}]')
                self.cancel_order(vt_orderid)
                canceled = True

        return canceled

    def has_no_activate_order(self, vt_symbol):
        """检查与当前 vt_symbol 相关的订单是否都已经被取消"""
        for vt_orderid in self.active_orderids.copy():
            if vt_orderid not in self.orders:
                continue
            order = self.orders[vt_orderid]
            if order.vt_symbol == vt_symbol:
                return False

        return True

    def set_targets_pos(self, targets: Dict[str, Tuple[int, Optional[float]]]):
        """设置目标仓位，目标价格（价格为None，则默认以最近的close下单）"""
        self.targets = targets
        self.target_refreshed = True

    def on_tick(self, tick: TickData) -> bool:
        """
        Callback of new tick data update.
        """
        is_available = super().on_tick(tick)
        if not is_available:
            return is_available
        # 更新 tick_arr
        tick_arr = self.tick_arr[tick.vt_symbol]
        tick_arr[:-1, :] = tick_arr[1:, :]
        tick_arr[-1, :] = [np.round(tick.last_price, 2), tick.volume]
        return is_available

    def on_bars(self, bars: Dict[str, BarData]) -> None:
        super().on_bars(bars)
        if not self._is_realtime_mode:
            # 实盘环境下通过tick进行交易触发
            # 回测环境下bar直接下单
            if self.handle_order_one_by_one:
                self._handle_order_to_target_pos_one_by_one_backtest_only()
            else:
                self._handle_order_to_target_pos_same_time_backtest_only(bars)

    def on_win_bars_finished(self, bars: Dict[str, BarData]) -> None:
        """对 on_win_bars 逻辑进行收尾处理"""
        super().on_win_bars_finished(bars)
        if not self._is_realtime_mode:
            # 实盘环境下通过tick进行交易触发
            # 回测环境下bar直接下单
            if self.handle_order_one_by_one:
                self._handle_order_to_target_pos_one_by_one_backtest_only()
            else:
                self._handle_order_to_target_pos_same_time_backtest_only(bars)

        self.put_event()

    def _handle_order_to_target_pos_same_time_backtest_only(self, bars: Dict[str, BarData]):
        """【仅用于回测】on_bar 触发套利下单交易"""
        # clear existed orders
        self.cancel_all()
        # Execute orders
        for vt_symbol in self.vt_symbols:
            target_pos, target_price = self.targets[vt_symbol]
            current_pos = self.get_pos(vt_symbol)

            pos_diff = target_pos - current_pos
            volume = abs(pos_diff)
            bar = bars[vt_symbol]
            if bar is None:
                self.write_log(f"{vt_symbol} {current_pos} -> {target_pos} but current bar is None")
                continue
            # datetime_str = datetime_2_str(bar.datetime)
            if pos_diff > 0:
                # self.write_log(
                #     f"{datetime_str} {vt_symbol} {current_pos} -> {target_pos} = {pos_diff} ↑ price={price}")

                if current_pos < 0:
                    self.cover(vt_symbol, target_price, volume)
                else:
                    self.buy(vt_symbol, target_price, volume)
            elif pos_diff < 0:
                # self.write_log(
                #     f"{datetime_str} {vt_symbol} {current_pos} -> {target_pos} = {pos_diff} ↓ price={price}")

                if current_pos > 0:
                    self.sell(vt_symbol, target_price, volume)
                else:
                    self.short(vt_symbol, target_price, volume)
            else:
                # self.write_log(
                #     f"{datetime_str} {vt_symbol} {current_pos} no change")
                pass

        self.target_refreshed = False

    def estimate_deal_seconds(self, vt_symbol):
        """评估指定价位成交所需花费的时间，按秒为单位计算
        -1 代表无法确定
        0 代表即时成交
        正数，代表预估相应秒左右成交
        """
        estimate_result = -1
        last_tick = self.bgs[vt_symbol].last_tick
        if last_tick is None:
            return estimate_result
        target_pos, target_price = self.targets[vt_symbol]
        target_price = np.round(target_price, 2)
        current_pos = self.get_pos(vt_symbol)
        tick_arr = self.tick_arr[vt_symbol]
        pos_diff = target_pos - current_pos
        if pos_diff > 0:
            # 做多或平空
            if last_tick.ask_price_1 <= target_price:
                # 目标价格在 卖一价 以上，即时成交
                estimate_result = 0
            elif target_price < last_tick.bid_price_1:
                # 目标价格在 买一价 以下，无法确定成交时间
                pass
            else:
                indexes = np.where(tick_arr[:, 0] == target_price)
                if len(indexes) > 0 and len(indexes[0]) > 0:
                    # 从第一个目标价格位置开始计算，所有目标价格的成交量平均到每秒的平均成交量
                    idx_first = indexes[0][0]
                    tick_passed = tick_arr.shape[0] - idx_first
                    avg_vol = tick_arr[tick_arr[:, 0] == target_price, 1].sum() / tick_passed * 2
                else:
                    # 没有对应目标价格的为成交量数据，以总体成交量为样本进行计算平均每秒成交量
                    # mean() 为每个tick的成交量，*2 为每秒成交量，
                    # 但是预期未来不是所有成交量都会这个价格上成交，因此再 / 2，因此，相当于 *2/2 没有操作
                    avg_vol = tick_arr[:, 1].mean()  # * 2 / 2

                estimate_result = last_tick.volume / avg_vol

        elif pos_diff < 0:
            # 做空或平多
            if target_price <= last_tick.bid_price_1:
                # 目标价格在 买一价 以下，即时成交
                estimate_result = 0
            elif last_tick.ask_price_1 < target_price:
                # 目标价格在 买一价 以下，无法确定成交时间
                pass
            else:
                indexes = np.where(tick_arr[:, 0] == target_price)
                if len(indexes) > 0 and len(indexes[0]) > 0:
                    # 从第一个目标价格位置开始计算，所有目标价格的成交量平均到每秒的平均成交量
                    idx_first = indexes[0][0]
                    tick_passed = tick_arr.shape[0] - idx_first
                    avg_vol = tick_arr[tick_arr[:, 0] == target_price, 1].sum() / tick_passed * 2
                else:
                    # 没有对应目标价格的为成交量数据，以总体成交量为样本进行计算平均每秒成交量
                    # mean() 为每个tick的成交量，*2 为每秒成交量，
                    # 但是预期未来不是所有成交量都会这个价格上成交，因此再 / 2，因此，相当于 *2/2 没有操作
                    avg_vol = tick_arr[:, 1].mean()  # * 2 / 2

                estimate_result = last_tick.volume / avg_vol
        else:
            pass

        return estimate_result

    def estimate_deal_seconds_all(self):
        """计算leg1 leg2成交时间"""
        leg1_ratio, dt = self.deal_difficult_ratio[self.vt_symbol1]
        if (datetime.now() - dt).seconds > 1:
            leg1_ratio = self.estimate_deal_seconds(self.vt_symbol1)
            self.deal_difficult_ratio[self.vt_symbol1] = (leg1_ratio, datetime.now())

        leg2_ratio, dt = self.deal_difficult_ratio[self.vt_symbol2]
        if (datetime.now() - dt).seconds > 1:
            leg2_ratio = self.estimate_deal_seconds(self.vt_symbol2)
            self.deal_difficult_ratio[self.vt_symbol2] = (leg2_ratio, datetime.now())

        return leg1_ratio, leg2_ratio

    def _estimate_deal_difficult_ratio_by_bar_backtest_only(self) -> Tuple[float, float]:
        """通过最近一个bar判断成交难度（仅用于回测）"""
        ratios = [-1, -1]
        for num, vt_symbol in enumerate([self.vt_symbol1, self.vt_symbol2]):
            target_pos, target_price = self.targets[vt_symbol]
            target_price = np.round(target_price, 2)
            current_pos = self.get_pos(vt_symbol)
            bar = self.current_bars[vt_symbol]
            pos_diff = target_pos - current_pos
            if pos_diff > 0:
                # 做多或平空
                # 当前bar 涨得越多，意味着越难成交
                ratio = -(bar.close_price - bar.open_price)
            elif pos_diff < 0:
                # 做空或平多
                # 当前bar 涨得越多，意味着越易成交
                ratio = bar.close_price - bar.open_price
            else:
                ratio = -1

            ratios[num] = ratio

        return ratios[0], ratios[1]

    def _handle_pos_2_target(self, vt_symbol, add_price=None, market_price=False, activate=True):
        """将制定合约持仓数调整到目标仓位"""
        target_pos, target_price = self.targets[vt_symbol]
        current_pos = self.get_pos(vt_symbol)

        if market_price:
            # 根据市场价格修正报单价格
            tick = self.bgs[vt_symbol].last_tick
            if target_pos - current_pos > 0:
                # 方向做多
                # activate==True 市价对手方主动成交，否则本方挂单成交
                target_price = tick.ask_price_1 if activate else tick.bid_price_1
            else:
                # 方向做空
                # activate==True 市价对手方主动成交，否则本方挂单成交
                target_price = tick.bid_price_1 if activate else tick.ask_price_1

            self.write_log(f"{vt_symbol} 盘口{'对价' if activate else '挂单'} {target_price:.1f}", 'debug')

        elif add_price is not None:
            if target_pos - current_pos > 0:
                # 方向做空
                target_price += add_price
            else:
                target_price -= add_price

            self.write_log(f"{vt_symbol} 对手方向加价挂单 {target_price:.1f}", 'debug')

        else:
            self.write_log(f"{vt_symbol} 限价挂单 {target_price:.1f}", 'debug')

        # 根据目标持仓以及当前持仓，规划下单方式
        if 0 <= target_pos < current_pos:
            # 平多单(减仓)
            volume = abs(target_pos - current_pos)
            self.sell(vt_symbol, target_price, volume)
        elif target_pos < 0 < current_pos:
            # 多翻空
            self.sell(vt_symbol, target_price, current_pos)
            # 先平仓，之后再开仓
            # self.short(vt_symbol, target_price, abs(target_pos))
        elif target_pos < current_pos <= 0:
            # 追加空单
            volume = abs(target_pos - current_pos)
            self.short(vt_symbol, target_price, volume)
        elif current_pos < target_pos <= 0:
            # 平空单(减仓)
            volume = abs(target_pos - current_pos)
            self.cover(vt_symbol, target_price, volume)
        elif current_pos < 0 < target_pos:
            # 空翻多
            self.cover(vt_symbol, target_price, abs(current_pos))
            # 先平仓，之后再开仓
            # self.buy(vt_symbol, target_price, target_pos)
        elif 0 <= current_pos < target_pos:
            # 追加多单
            volume = abs(target_pos - current_pos)
            self.buy(vt_symbol, target_price, volume)

    def _handle_active_order_by_bar_backtest_only(self, vt_symbol, add_price=1):
        """执行主动单（仅依赖于 bar 数据），仅用于回测使用"""
        if len(self.active_orderids) == 0:
            # 首次开仓
            self.write_log(f"建立配对交易 {vt_symbol} 优先配对。", 'debug')
            self._handle_pos_2_target(vt_symbol, add_price=add_price)
            self.target_refreshed = False

        else:
            # 有开仓，等待成交
            if self.target_refreshed:
                # 如果超时，则重新下单
                # 检查新下单的价格是否与就单子相同，如果相同则不用 撤销重下单，继续等待即可
                # 注意：这里 list(self.active_orderids) 其中 list() 不能去掉，
                #       因为内部会 cancel_order(order.orderid) 造成 active_orderids 变化会抛异常
                for active_orderid in list(self.active_orderids):
                    order = self.get_order(active_orderid)
                    if order is None:
                        continue
                    target_pos, target_price = self.targets[order.vt_symbol]
                    if order.price != target_price:
                        # 不要在订单取消后面直接下单，
                        # 因为 cancel_order 可能会失败，引发导致重复下单的情况。
                        # 订单取消成功后,通过 if len(self.active_orderids) == 0 状态进行重新下单
                        # 重新下单后 self.target_refreshed = False
                        # 发现 cancel_order(order.orderid) 总是找不到订单导致撤单失败
                        # self.cancel_order(order.orderid)
                        self.cancel_all()

    def _handle_order_to_target_pos_one_by_one_backtest_only(self):
        """【仅用于回测】on_bar 触发套利下单交易"""
        # clear existed orders
        self.cancel_all()
        # Execute orders
        # 首先分别计算 symbol1 symbol2 的仓位差
        vt_symbol = self.vt_symbol1
        target_pos1, target_price1 = self.targets[vt_symbol]
        current_pos1 = self.get_pos(vt_symbol)
        pos_diff1 = target_pos1 - current_pos1
        pos_diff1_abs = abs(pos_diff1)
        vt_symbol = self.vt_symbol2
        target_pos2, target_price2 = self.targets[vt_symbol]
        current_pos2 = self.get_pos(vt_symbol)
        pos_diff2 = target_pos2 - current_pos2
        pos_diff2_abs = abs(pos_diff2)

        if pos_diff1 == pos_diff2 == 0:
            # 没有需要操作的合约或订单
            return

        # 确定哪一条是主动腿，哪一条是被动腿。判断逻辑如下：
        deal_passive_order = True
        if pos_diff1_abs > pos_diff2_abs:
            # 如果 leg1 leg2 数量不匹配，即，存在单腿，以瘸腿的一个为主动腿，另一条为被动腿
            # leg_active 主动腿, leg_passive 被动腿
            # TODO: 目前仅考虑 1:1 配对交易，其他情况稍后覆盖
            leg_passive = self.vt_symbol2
            leg_active = self.vt_symbol1
            # 发主动成交申请
            deal_passive_order = False
            # 重置主动单等待时间
            self.active_order_waiting_since_dt = None
        elif pos_diff1_abs < pos_diff2_abs:
            leg_passive = self.vt_symbol1
            leg_active = self.vt_symbol2
            # 发主动成交申请
            deal_passive_order = False
            # 重置主动单等待时间
            self.active_order_waiting_since_dt = None
        else:
            # 不存在单腿情况，进行成交评估。
            leg_passive, leg_active = None, None
            if len(self.active_orderids) > 0:
                # self.get_all_active_orderids() 实际返回值是 list(self.active_orderids)
                # 因此 orderid 是 str
                for orderid in self.get_all_active_orderids():
                    order = self.get_order(orderid)
                    if order is None:
                        continue

                    leg_passive = order.vt_symbol
                    leg_active = self.vt_symbol1 if self.vt_symbol1 != order.vt_symbol else self.vt_symbol2
                    break

            if leg_passive is None:
                # 选择评估时间长的作为被动腿（优先下单），主动腿等被动腿成交后在发交易申请
                leg1_ratio, leg2_ratio = self._estimate_deal_difficult_ratio_by_bar_backtest_only()
                if leg1_ratio == leg2_ratio:
                    # 如果难度相同，leg2 先下单
                    leg_passive = self.vt_symbol2
                    leg_active = self.vt_symbol1
                # elif leg1_ratio == -1:
                #     # leg1 难成交，leg1 先下被动单
                #     leg_passive = self.vt_symbol1
                #     leg_active = self.vt_symbol2
                # elif leg2_ratio == -1:
                #     # leg2 难成交，leg2 先下被动单
                #     leg_passive = self.vt_symbol2
                #     leg_active = self.vt_symbol1
                elif leg1_ratio < leg2_ratio:
                    # leg1 成交耗时比 leg2 短，leg2 先下被动单
                    leg_passive = self.vt_symbol2
                    leg_active = self.vt_symbol1
                else:
                    # leg2 成交耗时比 leg1 短，leg1 先下被动单
                    leg_passive = self.vt_symbol1
                    leg_active = self.vt_symbol2

        if deal_passive_order:
            self._handle_passive_order(leg_passive)
        else:
            self._handle_active_order_by_bar_backtest_only(leg_active)

        self.target_refreshed = False


class TargetPosAndPriceSpreadTemplate(TargetPosAndPrice2SymbolTemplate):
    # 当前价差
    curr_spread = 0
    # spread array size
    spread_array_size = 100

    def __init__(
            self,
            strategy_engine: StrategyEngine,
            strategy_name: str,
            vt_symbols: List[str],
            setting: dict
    ):
        """"""
        super().__init__(strategy_engine, strategy_name, vt_symbols, setting)
        # 价差数组
        self.price_spreads = np.zeros(0)

    def on_init(self) -> None:
        super().on_init()
        self.price_spreads = np.zeros(self.spread_array_size)

    def update_spread(self, spread):
        self.price_spreads[:-1] = self.price_spreads[1:]
        self.price_spreads[-1] = self.curr_spread = spread


if __name__ == "__main__":
    pass
